//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __Spectrum__
#define __Spectrum__

#include <TMath.h>
#include <TGraph.h>
#include <TH1.h>
#include <TFile.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include "CUtils.h"
#include "FFT.h"
#
using namespace std;

/**
 * Compute data spectra.
 * This class was designed to compute the one-sided power spectrum of a given stretch of data (PSD). It uses the median-mean method described in <a href="http://arxiv.org/abs/gr-qc/0509116">gr-qc/0509116</a>. This method can be represented by the cartoon below:
 * \verbatim
 |--------------------------------| input data vector (time domain)
 
 |---|---|---|---|---|---|---|---|   Neven 'even' subsegments
   |---|---|---|---|---|---|---|     Nodd 'odd' subsegments
 \endverbatim
 *
 * The input data vector is divided into Neven 'even' subsegments and Nodd 'odd' subsegments. Odd and even subsegments overlap by 50%. Neven and Nodd are determined by the size of the input vector and the requested frequency resolution for the PSD. The minimum requirement is to have at least Neven=1 and Nodd=0 to estimate the PSD (not the best estimate!).
 * Each subsegment is hann-windowed and FFTed to get Nodd+Neven power spectra. The median PSDs are computed for odd and even subsegments separately and a median-to-average correction factor is applied. The 2 PSDs are finally averaged into one.
 *
 * In fact, a dynamical PSD estimation is in use. This means that number of even/odd sub-segments is defined and fixed by the constructor of this class. Then, the PSD can be updated by providing an arbitrary number of subsegment data: see AddData().
 *
 * The PSD is estimated with a given frequency resolution. A linear interpolation (in log) is used to compute the power for any given frequency.
 * \author    Florent Robinet
 */
class Spectrum{

 public:

  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the Spectrum class.
   * It defines the vector sizes and initiates the FFT plan (FFTW_MEASURE). The spectrum resolution is defined by the number of points used to compute the data spectrum (between 0 and the Nyquist frequency). A nominal duration over which the PSD is estimated must be given (see the method presented in the class description). From this a circular buffer of sub-segments is defined and will be used to apply the median-mean method.
   *
   * The data sampling frequency and the spectrum size must be even numbers. The nominal data duration must be long enough to contain at least 1 subsegment (1 even and 0 odd), in other words, aDataDuration*aDataSamplingFrequency >= 2*aSpectrumSize.
   *
   * The actual spectrum size is 'aSpectrumSize' + 1 to account for the DC frequency.
   * @param aSpectrumSize number of points in the spectrum
   * @param aDataDuration nominal data duration to estimate the PSD [s]
   * @param aDataSamplingFrequency input data sampling frequency [Hz]
   * @param averbose verbosity level
   */
  Spectrum(const int aSpectrumSize,
	   const int aDataDuration, 
	   const int aDataSamplingFrequency,
	   const int aVerbosity=0);

  /**
   * Destructor of the Spectrum class.
   */
  virtual ~Spectrum(void);
  /**
     @}
  */


  /**
   * Resets the current PSD and PSD circular buffer.
   * The PSD buffer is flushed. The current PSD is set to 0.
   */
  void Reset(void);
  
  /**
   * Loads a data vector and updates the current PSD.
   * The data vector 'aData' (time domain) is used to update the current PSD. The method presented in this class description is applied.
   *
   * The user must specify the size of the input vector to cover. It must be long enough to be able to compute at least 1 PSD (>=2*aSpectrumSize).
   *
   * If the size of the input vector is smaller than the size specified in the constructor, the circular buffer is updated with this new data set. The oldest PSDs in the buffer are erased. If the size of the input vector is larger than the size specified in the constructor, only the end of the input vector is considered (down to the size specified in the constructor). This case is computationnally inefficient as sub-segmenmt PSDs are computed for nothing.
   *
   * Optionally, the user can provide a starting index. The data vector will only be used starting at this index. The aDataSize value should then account for a shifted start (do not mess this up as it will not be checked!).
   * @param aDataSize input vector size (number of samples) starting at aDataStart
   * @param aData input data vector (time domain)
   * @param aDataStart first data index to consider for the PSD estimation (=0 if from start)
   */
  bool AddData(const int aDataSize, const double *aData, const int aDataStart=0);

  /**
   * Loads a data vector and computes a new PSD.
   * This function is the same as AddData() but the PSD circular buffer is flushed before loading the new data vector.
   * @param aDataSize input vector size (number of samples) starting at aDataStart
   * @param aData input data vector (time domain)
   * @param aDataStart first data index to consider for the PSD estimation (=0 if from start)
   */
  bool LoadData(const int aDataSize, const double *aData, const int aDataStart=0);

  /**
   * Returns the current power value at a given frequency.
   * The power value, is computed by logarithmically interpolating the PSD if the requested frequency is inside the PSD frequency range. If it is outside this range, the last known PSD value is returned.
   *
   * If the requested frequency is negative or zero, the power DC value is returned.
   *
   * -1.0 is returned if this function fails.
   *
   * If this function is called many times in a row, it is more optimal if the requested frequencies are called in a sorted order (increasing or decreasing).
   * @param aFrequency frequency value [Hz]
   */
  double GetPower(const double aFrequency);

  /**
   * Returns the current power value at a given frequency index.                                                                                       
   * @param aFrequencyIndex frequency index.
   * @pre The frequency index must be a valid index: from 0 to the value returned by GetSpectrumSize().
   */
  inline double GetPower(const int aFrequencyIndex){ return PSD[aFrequencyIndex]; };

  /**
   * Returns the current power variance value at a given frequency index.
   * @param aFrequencyIndex frequency index.
   * @pre The frequency index must be a valid index: from 0 to the value returned by GetSpectrumSize().
   */
  inline double GetPowerVariance(const int aFrequencyIndex){ return PSD_var[aFrequencyIndex]; };

  /**
   * Returns the current amplitude value at a given frequency.
   * The amplitude value, is computed by logarithmically interpolating the PSD if the requested frequency is inside the PSD frequency range and a square-root of the PSD is returned. If the PSD value is outside this range, the last known PSD value is used.
   *
   * If the requested frequency is negative or zero, the amplitude DC value (absolute value) is returned.
   *
   * -1.0 is returned if this function fails.
   *
   * If this function is called many times in a row, it is more optimal if the requested frequencies are called in a sorted order (increasing or decreasing).
   * @param aFrequency frequency value [Hz]
   */
  double GetAmplitude(const double aFrequency);

  /**
   * Returns the current PSD as a TGraph.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   */
  TGraph* GetPSD(void);

  /**
   * Returns the current PSD as a TGraph between 2 frequencies.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   * @param aFrequencyMin minimum frequency value [Hz]
   * @param aFrequencyMax minimum frequency value [Hz]
   */
  TGraph* GetPSD(const double aFrequencyMin, const double aFrequencyMax);

  /**
   * Returns the current ASD as a TGraph.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   */
  TGraph* GetASD(void);

  /**
   * Returns the current ASD as a TGraph.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   * @param aFrequencyMin minimum frequency value [Hz]
   * @param aFrequencyMax minimum frequency value [Hz]
   */
  TGraph* GetASD(const double aFrequencyMin, const double aFrequencyMax);

  /**
   * Returns the PSD for a given subsegment.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   * @param aParity 0 for even subsegments, 1 for odd subsegments
   * @param aIndex subsegment index
   */
  TGraph* GetSubPSD(const int aParity, const int aIndex);

  /**
   * Returns the ASD for a given subsegment.
   * The returned TGraph must be deleted by the user. NULL is return if this function fails.
   * @param aParity 0 for even subsegment, 1 for odd subsegment
   * @param aIndex subsegment index
   */
  TGraph* GetSubASD(const int aParity, const int aIndex);

  /**
   * Writes all sub-segment PSDs in a ROOT file.
   * @param aOutFileName ROOT output file name
   */
  bool WriteSubPSD(const string aOutFileName);

  /**
   * Writes all sub-segment ASDs in a ROOT file.
   * @param aOutFileName ROOT output file name
   */
  bool WriteSubASD(const string aOutFileName);

  /**
   * Returns the number of points in the spectrum.
   * The actual spectrum size is the returned value -1 to remove the DC frequency.
   */
  inline int GetSpectrumSize(void){ return fft_psd->GetSize_f(); };

  /**
   * Returns the frequency of a given spectrum frequency index.
   * @param aIndex frequency index (0=DC)
   */
  inline double GetSpectrumFrequency(const int aIndex){ return f_PSD[aIndex]; };

  /**
   * Returns the frequency resolution of the spectrum [Hz].
   */
  inline double GetSpectrumResolution(void){ return (double)DataSamplingFrequency/2.0/(double)fft_psd->GetSize_f(); };

  /**
   * Returns the Nyquist frequency of the spectrum [Hz].
   */
  inline int GetSpectrumNyquist(void){ return DataSamplingFrequency/2; };

  /**
   * Returns class status.
   */
  inline bool GetStatus(void){ return status_OK; };

  /**
   * Returns the maximum number of subsegments in the circular buffer.
   * @param aParity 0 for even subsegments, 1 for odd subsegments
   */
  inline int GetNSubSegmentsMax(const int aParity){
    return nSubSegments[aParity%2];
  };

  /**
   * Returns the effective data buffer length [s] used to compute the PSD.
   */
  inline int GetDataBufferLength(void){ return nSubSegments[0]*(fft_psd->GetSize_t()/DataSamplingFrequency); };

  /**
   * Checks the current status of the PSD buffer.
   * Returns true if the PSD buffer is empty, i.e. not a single PSD is available.
   */
  inline bool IsBufferEmpty(void){
    if(stop_buffer[0]<0&&stop_buffer[1]<0) return true;
    return false;
  };


 private:

  bool status_OK;              ///< class status
  int fVerbosity;              ///< verbosity level

  // input parameters
  int DataSamplingFrequency;   ///< data sampling frequency

  // derived parameters
  int nSubSegments[2];         ///< number of even/odd sub-subsegments

  // containers
  double *HannWindow;          ///< hann window vector
  double *logPSD;              ///< PSD vector
  double *PSD;                 ///< PSD vector
  double *PSD_var;             ///< PSD variance vector
  double *f_PSD;               ///< frequency vector
  double **PSD_buffer[2];      ///< buffer for individual PSDs
  bool   *flag_buffer[2];      ///< flag for individual PSDs
  int    stop_buffer[2];       ///< last filled buffer index

  fft *fft_psd;                ///< FFT plan to compute the PSD

  gsl_interp_accel *acc;       ///< interpolation accelerator
  gsl_spline *interp_psd;      ///< PSD interpolator
  
  double* GetHannWindow(const int aSize);///< create the hann window
  void ComputeMedianMean(void);///< compute median-mean PSD
  
  ClassDef(Spectrum,0)  
};

#endif


