//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "ReadTriggerMetaData.h"

ClassImp(ReadTriggerMetaData)

////////////////////////////////////////////////////////////////////////////////////
ReadTriggerMetaData::ReadTriggerMetaData(const string aPattern, const string aDirectory, const int aVerbose): ReadTriggerSegments (aPattern, aDirectory, aVerbose){ 
////////////////////////////////////////////////////////////////////////////////////
  
  // set to default
  Mfmin_stat   = 1e20;
  Mfmax_stat   = -1.0;
  Mqmin_stat   = 1e20;
  Mqmax_stat   = -1.0;
  Msnrmin_stat = 1e20;
  Msnrmax_stat = -1.0;
  Mstart       = NULL;
  Mend         = NULL;
  Mfmin        = NULL;
  Mfmax        = NULL;
  Mqmin        = NULL;
  Mqmax        = NULL;
  Msnrmin      = NULL;
  Msnrmax      = NULL;
  
  // get data (if present)
  if(status_OK){
    if(Verbose) cout<<"ReadTriggerMetaData::ReadTriggerMetaData: Get meta-data..."<<endl;
    status_OK*=GetInputMetaData();
  }
}

////////////////////////////////////////////////////////////////////////////////////
ReadTriggerMetaData::~ReadTriggerMetaData(void){
////////////////////////////////////////////////////////////////////////////////////
  if(Verbose>1) cout<<"ReadTriggerMetaData::~ReadTriggerMetaData"<<endl;
  if(Mstart!=NULL)    delete Mstart;
  if(Mend!=NULL)      delete Mend;
  if(Mfmin!=NULL)     delete Mfmin;
  if(Mfmax!=NULL)     delete Mfmax;
  if(Mqmin!=NULL)     delete Mqmin;
  if(Mqmax!=NULL)     delete Mqmax;
  if(Msnrmin!=NULL)   delete Msnrmin;
  if(Msnrmax!=NULL)   delete Msnrmax;
}

////////////////////////////////////////////////////////////////////////////////////
bool ReadTriggerMetaData::GetInputMetaData(void){ 
////////////////////////////////////////////////////////////////////////////////////

  // chain metadata trees...
  TChain *Mtree;
  if(Dir.compare(""))  Mtree = new TChain((Dir+"/metadata").c_str());
  else                 Mtree = new TChain("metadata");

  // get list of patterns and load it
  vector <string> patlist = SplitString(Pattern,' ');
  for(int p=0; p<(int)patlist.size(); p++) Mtree->Add(patlist[p].c_str());
  Mtree->SetName(("metadata_"+srandid).c_str());
  patlist.clear();

  if(!Mtree->GetNtrees()){
    if(Verbose) cout<<"ReadTriggerMetaData::GetInputMetaData: no meta-data"<<endl;  
  }
  
  // mandatory metadata: start / end
  double st;
  if(Mtree->GetBranchStatus("start")) Mtree->SetBranchAddress("start", &st);
  else{
    cerr<<"ReadTriggerMetaData::GetInputMetaData: missing 'start' field in metadata tree"<<endl;
    delete Mtree;
    return false;
  }
  double en;
  if(Mtree->GetBranchStatus("end")) Mtree->SetBranchAddress("end", &en);
  else{
    cerr<<"ReadTriggerMetaData::GetInputMetaData: missing 'end' field in metadata tree"<<endl;
    delete Mtree;
    return false;
  }

  // allocate memory
  Mstart = new double[Mtree->GetEntries()];
  Mend   = new double[Mtree->GetEntries()];
  
  // extract timing and compute live time
  double livetime=0;
  for(int m=0; m<Mtree->GetEntries(); m++){
    Mtree->GetEntry(m);
    livetime+=en-st;
    Mstart[m]=st;
    Mend[m]=en;
  }

  // check livetime against segments
  if(livetime < Seg->GetLiveTime()){// the meta-data set is incomplete
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: the meta-data set is not complete"<<endl;
    if(Verbose>2) cout<<"                                       segment livetime = "<<Seg->GetLiveTime()<<" while meta livetime = "<<livetime<<endl;
  }
  else if(livetime > Seg->GetLiveTime()){// should not happen
    cerr<<"ReadTriggerMetaData::GetInputMetaData: the meta-data set is incompatible with the segments"<<endl;
    delete Mtree;
    return false;
  }
  else{// the meta-data set is complete
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: the meta-data set is complete"<<endl;
  }
  
  // string branches
  string *pn = new string();
  string *sn = new string();

  // standard GWOLLUM metadata (optional)
  if(Mtree->GetBranchStatus("Mprocessname")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mprocessname field found"<<endl;
    Mtree->SetBranchAddress("Mprocessname", &pn);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      if(!t) Mprocessname=*pn;
      else{
	if(*pn!=Mprocessname){
	  cerr<<"ReadTriggerMetaData::GetInputMetaData: the process name should be unique"<<endl;
	  delete Mtree;
	  delete sn;
	  delete pn;
	  return false;
	}
      }
    }
  }

  if(Mtree->GetBranchStatus("Mstreamname")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mstreamname field found"<<endl;
    Mtree->SetBranchAddress("Mstreamname", &sn);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      if(!t) Mstreamname=*sn;
      else{
	if(*sn!=Mstreamname){
	  cerr<<"ReadTriggerMetaData::GetInputMetaData: the stream name should be unique"<<endl;
	  delete Mtree;
	  delete sn;
	  delete pn;
	  return false;
	}
      }
    }
  }

  double fmin;
  Mfmin = new double[Mtree->GetEntries()];
  if(Mtree->GetBranchStatus("Mfmin")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mfmin field found"<<endl;
    Mtree->SetBranchAddress("Mfmin", &fmin);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Mfmin[t]=fmin;
      if(fmin<0) continue;
      if(fmin<Mfmin_stat) Mfmin_stat=fmin;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Mfmin_stat = "<<Mfmin_stat<<endl;
  }

  double fmax;
  Mfmax = new double[Mtree->GetEntries()];
  if(Mtree->GetBranchStatus("Mfmax")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mfmax field found"<<endl;
    Mtree->SetBranchAddress("Mfmax", &fmax);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Mfmax[t]=fmax;
      if(fmax<0) continue;
      if(fmax>Mfmax_stat) Mfmax_stat=fmax;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Mfmax_stat = "<<Mfmax_stat<<endl;
  }

  double qmin;
  Mqmin = new double[Mtree->GetEntries()];  
  if(Mtree->GetBranchStatus("Mqmin")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mqmin field found"<<endl;
    Mtree->SetBranchAddress("Mqmin", &qmin);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Mqmin[t]=qmin;
      if(qmin<0) continue;
      if(qmin<Mqmin_stat) Mqmin_stat=qmin;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Mqmin_stat = "<<Mqmin_stat<<endl;
  }
  
  double qmax;
  Mqmax = new double[Mtree->GetEntries()];
  if(Mtree->GetBranchStatus("Mqmax")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Mqmax field found"<<endl;
    Mtree->SetBranchAddress("Mqmax", &qmax);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Mqmax[t]=qmax;
      if(qmax<0) continue;
      if(qmax>Mqmax_stat) Mqmax_stat=qmax;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Mqmax_stat = "<<Mqmax_stat<<endl;
  }
  
  double smin;
  Msnrmin = new double[Mtree->GetEntries()];
  if(Mtree->GetBranchStatus("Msnrmin")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Msnrmin field found"<<endl;
    Mtree->SetBranchAddress("Msnrmin", &smin);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Msnrmin[t]=smin;
     if(smin<0) continue;
      if(smin<Msnrmin_stat) Msnrmin_stat=smin;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Msnrmin_stat = "<<Msnrmin_stat<<endl;
  }
  
  double smax;
  Msnrmax = new double[Mtree->GetEntries()];
  if(Mtree->GetBranchStatus("Msnrmax")){
    if(Verbose>1) cout<<"ReadTriggerMetaData::GetInputMetaData: Msnrmax field found"<<endl;
    Mtree->SetBranchAddress("Msnrmax", &smax);
    for(int t=0; t<(int)Mtree->GetEntries(); t++){
      Mtree->GetEntry(t);
      Msnrmax[t]=smax;
      if(smax<0) continue;
      if(smax>Msnrmax_stat) Msnrmax_stat=smax;
    }
    if(Verbose>2) cout<<"ReadTriggerMetaData::GetInputMetaData: Msnrmax_stat = "<<Msnrmax_stat<<endl;
  }

  delete sn;
  delete pn;
  delete Mtree;
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
string ReadTriggerMetaData::Merge(const string aOutdir){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ReadTriggerMetaData::Merge: the ReadTriggerMetaData object is corrupted"<<endl;
    return "none";
  }
  if(!IsDirectory(aOutdir)){
    cerr<<"ReadTriggerMetaData::Merge: the directory "<<aOutdir<<" does not exist"<<endl;
    return "none";
  }

  //  get segment tree
  TTree *segtree = Seg->GetTree();
  TChain *triggers_tmp = new TChain("triggers");
  triggers_tmp->Add(Pattern.c_str());

  // output file name (GWOLLUM convention)
  ostringstream outfilestream;
  outfilestream<<aOutdir<<"/"<<Mstreamname<<"_"<<(int)floor(Seg->GetStart(0))<<"_"<<(int)ceil(Seg->GetEnd(Seg->GetNsegments()-1))-(int)floor(Seg->GetStart(0))<<".root";
  
  TFile* f = TFile::Open(outfilestream.str().c_str(), "RECREATE");
  f->cd();
  
  // segments
  segtree->Write();
  delete segtree;
   
  // triggers
  if(triggers_tmp->GetEntries()){
    TTree *triggers_copy = triggers_tmp->CloneTree(0);
    triggers_copy->SetName("triggers");
    triggers_copy->CopyEntries(triggers_tmp);
    triggers_copy->Write();
    delete triggers_copy;
  }
  else{
    // FIXME: this is ugly...
    TTree* t = new TTree("triggers", "triggers");
    double v1, v2, v3, v4, v5, v6, v7, v8, v9, v10;
    t->Branch("time",     &v1, "time/D");
    t->Branch("frequency",&v2, "frequency/D");
    t->Branch("q",        &v3, "q/D");
    t->Branch("snr",      &v4, "snr/D");
    t->Branch("tstart",   &v5, "tstart/D");
    t->Branch("tend",     &v6, "tend/D");
    t->Branch("fstart",   &v7, "fstart/D");
    t->Branch("fend",     &v8, "fend/D");
    t->Branch("amplitude",&v9, "amplitude/D");
    t->Branch("phase",    &v10,"phase/D");
    t->Write();
    delete t;
  }
  delete triggers_tmp;

  // metadata
  TChain *meta_tmp = new TChain("metadata");
  meta_tmp->Add(Pattern.c_str());
  TTree *meta_copy = meta_tmp->CloneTree(0);
  meta_copy->SetName("metadata");
  meta_copy->CopyEntries(meta_tmp);
  meta_copy->Write();
  delete meta_copy;
  delete meta_tmp;
  
  f->Close();

  return outfilestream.str();
}
